## ./japan.R
## 
## Produces main estimation results (Table 2)
## Designed to be run under LSF, or (with minor modifications) PBS
## To obtain parameter estimates, run without any arguments:

# bsub -n 8,20 -R "span[hosts=1]" -W 24:00 Rscript ./japan.R

## To obtain confidence intervals, submit the search grid as the array indices:

# bsub -n 8,20 -R "span[hosts=1]" -W 24:00 -J japan[6-10] Rscript ./japan.R CS B0

## (the array id becomes fixed.val.array.id, below)

## To check coverage probabilities, submit the bootstrap iteration as the array index:

# bsub -n 8,20 -R "span[hosts=1]" -W 24:00 Rscript ./japan.R boot 42

## For given parameters, pen(par) returns the test statistic, T.
## The return value is relative to 95th percentile.
## (i.e. parameters resulting in values above 1 should be rejected)

TAX.FLOOR <- TRUE  # Gives results in column I; FALSE will give results in column II 

library("Rcplex")
library("cplexAPI")
library("igraph") # ONE-BASED indexing
library("parallel")
library("censReg")
library("mvtnorm")
library("numDeriv")
library("combinat")

proj.home <- "/home/fas/weese/egw5/japan_qe/"
proj.tmp <- "/home/fas/weese/egw5/scratch/japan_qe/"

options(echo=TRUE)

## number of cores to use for child processes (ncpus - 1 seems to work well)
## options(mc.cores=7)
CORES <- as.numeric(Sys.getenv("LSB_DJOB_NUMPROC")) - 1
if(is.na(CORES) || CORES < 2) CORES <- 1
options(mc.cores=CORES)

set.seed(3)  # reproducibility

CPLEX.TILIM <- 300 # maximum time for a quadratic program (there will be problems if this binds anywhere close to the points we're interested in)
MAX.REFS <- 1000
SIM.LENGTH <- 1000 # takes about 50% more time per iteration to do 1000 than 100, but there seems to be a fair amount of simulation noise...
TEST.DRAWS <- 10000
SMOOTHING <- 0.1  # 0.01 creates almost no distortions, but may not be quite as good for finding minima
USE.TAX <- TRUE


par.calibrated <- c(B0=NA, B1=NA, B2=NA, B3=1, B4=1)
par.calibrated["tau.bar"] <- 0.084 * 4/3
par.calibrated["r"] <- 0.02
par.calibrated["disc.santei.h13"] <- par.calibrated["r"] * sum(c(1/(1+par.calibrated["r"])^(1:5), c(.9, .7, .5, .3, .1)/(1+par.calibrated["r"])^(6:10)))
par.calibrated["disc.santei.h20"] <- par.calibrated["r"] * sum(c(1/(1+par.calibrated["r"])^(1:10), c(.9, .7, .5, .3, .1)/(1+par.calibrated["r"])^(11:15)))
par.calibrated["disc.tokurei"] <- par.calibrated["r"] * sum(1/(1+par.calibrated["r"])^(1:10)) * 0.95 * 0.70 * (sum(1/(1+par.calibrated["r"])^(1:20))/20) / 10

par.fixed <- par.calibrated


### Process arguments, if there are any
tmp.args <- commandArgs(trailingOnly=TRUE)
if(is.na(tmp.args[1])) {
  print("called with no arguments")
  tmp.args <- "opt"
} else {
  print(paste("called with args", paste(tmp.args, collapse=",")))
  stopifnot(tmp.args[1] %in% c("opt", "CS", "boot"))
}
if(tmp.args[1] == "opt") {
  fixed.var <- NA
  boot <- NA
  print("straight optimization, all parameters free")
} else if(tmp.args[1] == "CS") {  # then we're checking to see whether there are any points in the confidence set with fixed.var set to fixed.val
  boot <- NA
  fixed.var <- tmp.args[2]
  stopifnot(fixed.var %in% c("B0", "B1", "B2", "B3", "B4"))
  fixed.val.arrayid <- as.numeric(tmp.args[3])  # intended to be an LSF array id, so following is set up to deal with this being a small integer
  if(is.na(fixed.val.arrayid)) {
      fixed.val.arrayid <- as.numeric(Sys.getenv("LSB_JOBINDEX"))
  }
  if(is.na(fixed.val.arrayid))
      stop("confidence set bound requires array index")
  print(paste("confidence set bound:", fixed.var, "will be fixed based on array id value", fixed.val.arrayid))
  if(fixed.var == "B0") par.fixed["B0"] <- fixed.val.arrayid  # 145 to 454  (39 - 348)
  if(fixed.var == "B1") par.fixed["B1"] <- fixed.val.arrayid / 10  # 26/10 to 121/10  (0/10 - 98/10)
  if(fixed.var == "B2") par.fixed["B2"] <- fixed.val.arrayid / 100 # 4/100 to 32/100   (0/100 - 39/100)
  if(fixed.var == "B3") par.fixed["B3"] <- fixed.val.arrayid / 100 # 95/100 to 112/100  (97/100 - 125/100)
  if(fixed.var == "B4") par.fixed["B4"] <- fixed.val.arrayid / 100 # 38/100 to 60/100  (29/100 - 66/100)
  print(paste("using", fixed.var, "=", par.fixed[fixed.var]))
} else if(tmp.args[1] == "boot") { # then we're doing a bootstrap
  fixed.var <- NA
  boot <- as.integer(tmp.args[2])
  if(is.na(boot)) stop("boot requires integer argument")
  set.seed(boot)
  print(paste("bootstrap: iteration", boot))
}

   
prefs <- 1:47
prefs

fname <- paste(proj.home, "cplex_status.R", sep="")
source(fname)

fname <- paste(proj.home, "h13.RData", sep="")
load(fname)
rownames(d.h13) <- as.character(as.numeric(rownames(d.h13)))

fname <- paste(proj.home, "pgraph.RData", sep="")
load(fname)

fname <- paste(proj.home, "v.R", sep="")
source(fname)

fname <- paste(proj.home, "soumu_choukakazei.txt", sep="")
tmp <- read.table(file(fname, encoding="UTF-16"), header=TRUE, row.names=3, sep="\t")
rownames(tmp) <- substring(rownames(tmp), first=1, last=nchar(rownames(tmp)) - 1)
tax.h22 <- tmp[tmp[,"団体名"]!="東京都",]

fname <- paste(proj.home, "gappei.RData", sep="")
load(fname)

singleton.dat.l <- list()
for(p in 1:47) {
  fname <- paste(proj.tmp, p, "/i_long_1_m_new.RData", sep="")
  load(fname)
  rownames(m.dat) <- V(p.graph.l[[p]])$name
  singleton.dat.l[[p]] <- m.dat
}

cnames <- c(colnames(tax.h22)[-(2:3)], colnames(m.dat))
actual.tax <- matrix(NA, nrow=nrow(tax.h22), ncol=length(cnames), dimnames=list(rownames(tax.h22), cnames))
actual.tax[,colnames(tax.h22)[-(2:3)]] <- as.matrix(tax.h22[,-(2:3)])

for(jis in rownames(tax.h22)) {
  p <- tax.h22[jis,"都道府県コード"]
  jis.use <- jis
  if(as.numeric(jis) %in% as.numeric(changes[,"jis.new"])) {
    jis.use <- changes[which(as.numeric(jis) == as.numeric(changes[,"jis.new"])),"jis.old"] } else {
      if (jis == "14150") jis.use <- "14209"
      if (jis == "15100") jis.use <- "15201"
      if (jis == "22130") jis.use <- "22202"
      if (jis == "27140") jis.use <- "27201"
      if (jis == "33100") jis.use <- "33201"
    }
  if(as.numeric(jis.use)%in% as.numeric(unlist(gappei[,"jis"]))) {
      if(p < 10) fname <- paste(proj.tmp, p, "/r0", jis.use, "_new.RData", sep="")
      if(p >= 10) fname <- paste(proj.tmp, p, "/r", jis.use, "_new.RData", sep="")
    load(fname)
    actual.tax[jis,colnames(m.dat)] <- dat.l[[length(dat.l)]][1,colnames(m.dat)]
   } else {
     actual.tax[jis,colnames(m.dat)] <- singleton.dat.l[[p]][jis.use,]
   }
}



vn <- function(n)
  paste("v", n, sep="")

mn <- function(n) paste("m", n, sep="")

mdn <- function(n) {
  mat.tmp <- outer(n, n, paste, sep="_")
  tmp <- mat.tmp[upper.tri(mat.tmp)]
  ret <- paste("md", tmp, sep="")
  return(ret)
}



make.p.r <- function(p) {
  r.l <- list()  # (m216_217 may need to be fixed manually)

  if(is.na(boot)) qc.dir <- paste(proj.tmp, p, sep="")
  if(!is.na(boot)) qc.dir <- paste(proj.tmp, p, "/boot/", boot, "/", sep="")

  refinements <- list.files(qc.dir, pattern="r.*_new.RData", full.names=TRUE)

  p.graph <- p.graph.l[[p]]

  for(refinement in refinements) {
    r.l[[refinement]] <- new.env()
    load(refinement, envir=r.l[[refinement]])
    if(length(r.l[[refinement]]$muni.l) > MAX.REFS) { # then we need to cut some...
      singleton.end <- max(which(lapply(r.l[[refinement]]$muni.l, length) == 1))
      actual <- length(r.l[[refinement]]$muni.l)
      tmp <- sort(sample((singleton.end+1):(actual-1), MAX.REFS - (singleton.end+1)))
      selected <- c(1:singleton.end, tmp, actual)
      r.l[[refinement]]$muni.l <- r.l[[refinement]]$muni.l[selected]
      r.l[[refinement]]$dat.l <- r.l[[refinement]]$dat.l[selected]
    }
    r.l[[refinement]]$muni.jis.l <- lapply(r.l[[refinement]]$muni.l, function(x) V(p.graph)$name[x + 1])
  }

  do.one <- function(r.e) {
    munis.jis <- r.e$muni.jis.l[[length(r.e$muni.jis.l)]]
    edgem.use.l <- list()
    p.graph <- p.graph.l[[p]]
    edgel <- E(p.graph)
    edgem <- t(sapply(as.list(1:length(edgel)), function(x) V(p.graph)[inc(edgel[x])]$name))
    in.merger <- function(edge.munis) all(edge.munis %in% munis.jis)
    use.edge <- apply(edgem, 1, in.merger)
    edgem.use <- edgem[use.edge,,drop=FALSE]
    rm(p.graph)
    cnames <- c(vn(munis.jis), mdn(munis.jis))
    rnames <- paste("d", 1:(length(r.e$muni.l) - 1), sep="")
    A <- matrix(0, nrow=length(rnames), ncol=length(cnames), dimnames=list(rnames, cnames))
    for(dev in 1:(length(r.e$muni.l)-1)) {
      dev.jis <- r.e$muni.jis.l[[dev]]
      A[dev, vn(munis.jis)] <- sqrt(sum(d.h13[munis.jis, "pop"])) * d.h13[munis.jis, "pop"] / sum(d.h13[munis.jis, "pop"])
      A[dev, vn(dev.jis)] <- A[dev, vn(dev.jis)] - sqrt(sum(d.h13[dev.jis, "pop"])) * d.h13[dev.jis, "pop"] / sum(d.h13[dev.jis, "pop"])
      mat.tmp <- outer(d.h13[munis.jis, "pop"], d.h13[munis.jis, "pop"])
      A[dev, mdn(munis.jis)] <- mat.tmp[upper.tri(mat.tmp)] / sum(d.h13[munis.jis, "pop"])^2
      if(length(dev.jis) > 1) {
        mat.tmp <- outer(d.h13[dev.jis, "pop"], d.h13[dev.jis, "pop"])
        A[dev, mdn(dev.jis)] <- A[dev, mdn(dev.jis)] - mat.tmp[upper.tri(mat.tmp)] / sum(d.h13[dev.jis, "pop"])^2
      }
    }
    bvec <- rep(NA, length(rnames))
    names(bvec) <- rnames
    cvec <- rep(0, length(cnames))
    names(cvec) <- cnames
    lb <- rep(NA, length(cnames))
    names(lb) <- cnames
    lb[vn(munis.jis)] <- -Inf
    lb[mdn(munis.jis)] <- 0
    sense <- rep("L", length(rnames))
    names(sense) <- rnames
    md.edges.tmp <- apply(edgem.use, 1, paste, collapse="_")
    md.edges <- paste("md", md.edges.tmp, sep="")
    stopifnot(md.edges %in% cnames)
    edge.pops <- apply(edgem.use, 1:2, function(i) d.h13[i, "pop"])
    edge.weights <- apply(edge.pops, 1, prod)/apply(edge.pops, 1, sum)
    Q <- matrix(0, nrow=length(cnames), ncol=length(cnames), dimnames=list(cnames, cnames))
    diag(Q)[vn(munis.jis)] <- d.h13[munis.jis, "pop"]
    function(b, output) {
      v.actual <- v.m.new(d=r.e$dat.l[[length(r.e$dat.l)]], b=b, time=1)
      names(v.actual) <- munis.jis
      if(any(v.actual==-Inf)) return(Inf)
      g <- function(muni.jis, dat) {
        tmp <- v.actual[muni.jis] - v.m.new(dat, b=b, time=1)
        return(max(tmp))
      }
      v.diff <- mapply(g, r.e$muni.jis.l[-length(r.e$muni.jis.l)], r.e$dat.l[-length(r.e$dat.l)])
      if(any(v.diff==-Inf)) return(Inf)
      stopifnot(length(v.diff) == length(rnames))
      bvec <- pmin(v.diff, .Machine$integer.max) # try to deal with some of the Infs
      cplex.out <- Rcplex(cvec=cvec, Amat=A, bvec=bvec, Qmat=Q, lb=lb, sense=sense, objsense="min", control=list(tilim=CPLEX.TILIM, trace=output))
      names(cplex.out$xopt) <- cnames
      names(cplex.out$extra$lambda) <- rnames
      if(!(cplex.out$status %in% cplex.status.ok)) {
        violations <- A %*% cplex.out$xopt > bvec
        if(any(violations)) {
          v.max <- max(cplex.out$extra$slack[violations])
          v.str <- paste(":  violation by", v.max)
        } else v.str <- ""
        print(paste(cplex.status[cplex.out$status], v.str, sep=""))
      }
      return(cplex.out$obj)
    }
  }
  f.l <- lapply(r.l, do.one)
  function(b, output=0) {
    tmp <- lapply(f.l, do.call, args=list(b=b, output=output))
    ret <- sum(unlist(tmp))
    return(ret)
  }
}


make.p.c <- function(p) {  #needed to split this off from above to avoid some Encode related bug...
    if(is.na(boot)) qc.dir <- paste(proj.tmp, p, "/", sep="")
    if(!is.na(boot)) qc.dir <- paste(proj.tmp, p, "/boot/", boot, "/", sep="")
  c.l <- list()
  coarsenings <- system(paste("ls ", qc.dir, "c*_new.RData", sep=""), intern=TRUE)
  for(coarsening in coarsenings) {
    e <- new.env()
    load(coarsening, envir=e)
    if((length(e$dat.l) == 3) && all(sapply(e$muni.l, length) == c(1, 1, 2)))
      c.l[[coarsening]] <- e
  }
  if(length(c.l) == 0) return(function(b, output=0) return(list(list(obj=0))))  # (prefecture 32 really doesn't have any singleton-singleton deviations)
  munis <- sort(unique(c(sapply(c.l, function(x) unlist(x$muni.l)))))
  muni.pairs <- sapply(c.l, function(x) x$muni.l[[3]])
  stopifnot(munis == sort(unique(c(muni.pairs))))

  tmp <- apply(muni.pairs, 2, paste, collapse="_")

  cnames <- c(vn(munis), mn(munis), paste("md", tmp, sep=""), c(outer(c("a", "b"), tmp, paste, sep="")))

  rnames.tmp <- c(sapply(as.list(tmp), function(x) paste("d", x, c("a", "b"), sep="")))

  rnames <- c(rnames.tmp, paste("B", tmp, sep=""))

  A <- matrix(0, nrow=length(rnames), ncol=length(cnames), dimnames=list(rnames, cnames))

  pop.munis <- rep(NA, max(muni.pairs) + 1)

  bvec <- c(rep(NA, length(muni.pairs)), rep(0, ncol(muni.pairs)))
  names(bvec) <- rnames
  cvec <- c(rep(0, 2*length(munis)), rep(1, ncol(muni.pairs)), rep(0, 2*ncol(muni.pairs)))
  lb <- c(rep(-Inf, length(munis)), rep(0, length(munis) + 3*ncol(muni.pairs)))

  sense <- c(rep("G", length(muni.pairs)), rep("E", ncol(muni.pairs)))
  vtype <- c(rep("C", 2*length(munis) + ncol(muni.pairs)), rep("B", 2*ncol(muni.pairs)))
  use.rows <- c(rep(FALSE, length(muni.pairs)), rep(TRUE, ncol(muni.pairs)))
  names(use.rows) <- rnames

  for(m in 1:ncol(muni.pairs)) {
    pair <- muni.pairs[,m]
    pair.str <- paste(pair, collapse="_")
    A[paste("B", pair.str, sep=""), c(paste("a", pair.str, sep=""), paste("b", pair.str, sep=""))] <- 1
    pop.a <- c.l[[m]]$dat.l[[1]][1,"pop"]
    pop.b <- c.l[[m]]$dat.l[[2]][1,"pop"]
    pop.munis[pair+1] <- c(pop.a, pop.b)
    pop.merger <- pop.a + pop.b

    A[paste("d", pair.str, "a", sep=""), paste("v", pair[1], sep="")] <- sqrt(pop.merger) * pop.a/pop.merger - sqrt(pop.a)
    A[paste("d", pair.str, "a", sep=""), paste("v", pair[2], sep="")] <- sqrt(pop.merger) * pop.b/pop.merger
    A[paste("d", pair.str, "a", sep=""), paste("md", pair.str, sep="")] <- 1

    A[paste("d", pair.str, "b", sep=""), paste("v", pair[1], sep="")] <- sqrt(pop.merger) * pop.a/pop.merger
    A[paste("d", pair.str, "b", sep=""), paste("v", pair[2], sep="")] <- sqrt(pop.merger) * pop.b/pop.merger - sqrt(pop.b)
    A[paste("d", pair.str, "b", sep=""), paste("md", pair.str, sep="")] <- 1
  }

  get.obj <- function(b, output=0) {
    for(m in 1:ncol(muni.pairs)) {
      tmp.actual <- c(v.m.new(c.l[[m]]$dat.l[[1]], b), v.m.new(c.l[[m]]$dat.l[[2]], b))
      tmp.merger <- v.m.new(c.l[[m]]$dat.l[[3]], b)
      tmp <- tmp.merger - tmp.actual
      a.inf <- tmp.actual[1] == -Inf
      b.inf <- tmp.actual[2] == -Inf
      if(a.inf && b.inf) return(list(list(obj=Inf)))  # the coarsening is unavoidable, so these are bad parameters

      bvec[m*2 - 1] <- if(a.inf) 0 else tmp[1]
      bvec[m*2] <- if(b.inf) 0 else tmp[2]
      if(!a.inf) use.rows[paste("d", pair.str, "a", sep="")] <- TRUE
      if(!b.inf) use.rows[paste("d", pair.str, "b", sep="")] <- TRUE
      if(!a.inf && !b.inf) {
        M <- max(10, bvec[c(m*2 - 1, m*2)]) * 2  # big M
        A[paste("d", pair.str, "a", sep=""), paste("a", pair.str, sep="")] <- M
        A[paste("d", pair.str, "b", sep=""), paste("b", pair.str, sep="")] <- M
        bvec[paste("B", pair.str, sep="")] <- 1
      }
    }
    stopifnot(length(munis) == sum(!is.na(pop.munis)))
    Q <- diag(x=c(na.omit(pop.munis), rep(0, length(munis)), rep(0, ncol(muni.pairs)), rep(0,2*ncol(muni.pairs))))

    cplex.out <- Rcplex(cvec=cvec, Amat=A[use.rows,], bvec=bvec[use.rows], Qmat=Q, lb=lb, sense=sense[use.rows], vtype=vtype, control=list(round=TRUE, tilim=CPLEX.TILIM, trace=output))

    if(!(cplex.out$status %in% cplex.status.ok)) print(cplex.status[cplex.out$status])
    names(cplex.out$xopt) <- cnames
    return(list(cplex.out))
  }

  return(get.obj)
}



make.p.m <- function(p) {
  p.graph <- p.graph.l[[p]]
  qc.dir <- paste(proj.tmp, p, "/", sep="")

  fname <- paste(qc.dir, "m2_nc.RData", sep="")
  load(fname)

  use.e <- apply(pairs.jis, 1, function(pair) are.connected(p.graph, V(p.graph)[pair[1]], V(p.graph)[pair[2]]))
  pairs.jis <- pairs.jis[use.e,]
  pairs.dat <- pairs.dat[use.e]
  
  pop.jis <- sapply(singles.dat, function(x) x[,"pop"])
  names(pop.jis) <- V(p.graph)$name

  mean.q <- matrix(rnorm(SIM.LENGTH*vcount(p.graph)), nrow=vcount(p.graph), ncol=SIM.LENGTH)
  rownames(mean.q) <- V(p.graph)$name
  mean <- mean.q / sqrt(pop.jis)
  
  var.q <- matrix(rnorm(SIM.LENGTH*vcount(p.graph)), nrow=vcount(p.graph), ncol=SIM.LENGTH)
  rownames(var.q) <- V(p.graph)$name
  var <- var.q * sqrt(2) / sqrt(pop.jis - 1)
  
  get.var <- function(i, j)
    (var[i,] * pop.jis[i] + var[j,] * pop.jis[j] + (mean[i,] - mean[j,])^2 * pop.jis[i] * pop.jis[j]/(pop.jis[i] + pop.jis[j]))/(pop.jis[i] + pop.jis[j])
 
  pairs.var <- t(apply(pairs.jis, 1, function(x) get.var(x[1], x[2])))
  pairs.eps <- sqrt(apply(pairs.jis, 1, function(x) sum(pop.jis[x]))) * pairs.var

  singles.eps <- sqrt(pop.jis) * var
  
  tmp <- simplify2array(pairs.dat)
  pairs.dat.a <- aperm(tmp, c(1,3,2))

  env <- openEnvCPLEX()
  prob <- initProbCPLEX(env)
  
  cn <- V(p.graph)$name
  rn <- paste("c", 1:nrow(pairs.jis), sep="")
  nc <- length(cn)
  nr <- length(rn)

  rhs <- rep(0, nr)
  obj <- rep(1, nc)
  sense <- rep("G", nr)
  lb <- rep(0, nc)
  ub <- rep(1, nc)
  
  tmp <- sapply(as.list(singles.jis), function(jis) apply(pairs.jis, 1, function(pair) jis %in% pair))
  tmp.idx <- apply(tmp, 2, which)

  cnt <- sapply(tmp.idx, length)
  ind <- as.vector(unlist(tmp.idx)) - 1

  beg <- cumsum(c(0, cnt[-length(cnt)]))

  val <- rep(1, nr*2)
  stopifnot(checkCopyLpwNamesCPLEX(env, prob, nc, nr, CPX_MIN, obj, rhs, sense, beg, cnt, ind, val, lb, ub, NULL, cn, rn) == 0)
  copyLpwNamesCPLEX(env, prob, nc, nr, CPX_MIN, obj, rhs, sense, beg, cnt, ind, val, lb, ub, NULL, cn, rn)

  xctype <- rep(CPX_CONTINUOUS, nc)
  stopifnot(checkCopyColTypeCPLEX(env, prob, xctype) == 0)
  copyColTypeCPLEX(env, prob, xctype)
  
  get.obj <- function(b, time) {
    v.single <- sapply(singles.dat, v.m.new, b=b, time=time)
    v.pair <- apply(pairs.dat.a, 1, v.m.new, b=b, time=time)

    u.single <- v.single - singles.eps
    u.single.pairm.tmp <- u.single[pairs.jis,]
    u.single.pairm.a <- array(u.single.pairm.tmp, c(dim(v.pair), SIM.LENGTH))
    
    u.pair.tmp <- sapply(list(v.pair[,1], v.pair[,2]), function(v) v - pairs.eps, simplify="array")
    u.pair.a <- aperm(u.pair.tmp, c(1,3,2))

    u.diff.pos <- pnorm(u.pair.a[,1,] - u.single.pairm.a[,1,], sd=SMOOTHING) * pnorm(u.pair.a[,2,] - u.single.pairm.a[,2,], sd=SMOOTHING)

    do.cplex <- function(rhs) {
      chgRhsCPLEX(env, prob, nrow(pairs.jis), (1:nrow(pairs.jis))-1, as.numeric(rhs)) -> rhs.chk
      stopifnot(rhs.chk==0)
      mipoptCPLEX(env, prob) -> mip.chk
      stopifnot(mip.chk==0)
      cplex.out <- solutionCPLEX(env, prob)
      return(cplex.out)
    }
    
    tmp.out <- apply(u.diff.pos, 2, do.cplex)
    ret <- sapply(tmp.out, function(x) x$objval)
    return(ret)
  }

  return(get.obj)
}


nmunis <- sapply(1:47, function(p) vcount(p.graph.l[[p]]))

fname <- paste(proj.home, "H9.RData", sep="")
load(fname)
dimnames(cross)[[1]] <- substring(dimnames(cross)[[1]], first=1, last=5)
ritou.tmp <- general[,37]
names(ritou.tmp) <- substring(general[,3], first=1, last=5)
ritou <- as.character(as.numeric(names(which(ritou.tmp==1))))

if(is.na(boot)) {
  fname <- paste(proj.home, "early_mergers.R", sep="")
  source(fname)
} else {
  early <- matrix(0, nrow=12, ncol=47, dimnames=list(1979:1990, names(p.graph.l)))
  for(p in 1:47) {
    fname <- paste(proj.tmp, p, "/boot/", boot, "-0.RData", sep="")
    load(fname)
    early[1,p] <- length(unlist(ret.l.0))
  }
  late <- vector("numeric", 47)
  for(p in 1:47) {
    fname <- paste(proj.tmp, p, "/boot/", boot, "-1.RData", sep="")
    load(fname)
    late[p] <- length(unlist(ret.l.1))
  }
}

pop.order <- rep(NA, 47)

pop.order[order(sapply(p.graph.l, function(p.graph) mean(d.h13[setdiff(V(p.graph)$name, ritou), "pop"] < 10000)))] <- 1:47
names(pop.order) <- names(p.graph.l)

pref.groups <- list()
pref.groups[[1]] <- intersect(prefs, which(pop.order <= 3))
pref.groups[[2]] <- intersect(prefs, which(pop.order > 3 & pop.order <= 39))
pref.groups[[3]] <- intersect(prefs, which(pop.order > 39))

stopifnot(length(unlist(pref.groups)) == length(prefs))

f.r.l <- mclapply(prefs, make.p.r)
names(f.r.l) <- as.character(prefs)

f.c.l <- mclapply(prefs, make.p.c)
names(f.c.l) <- as.character(prefs)

f.m.l <- lapply(prefs, make.p.m)
names(f.m.l) <- as.character(prefs)


pen.tax <- function(pnums, b) {
  rows.use <- actual.tax[,"都道府県コード"] %in% pnums
  actual.chou <- b["tau.bar"] * actual.tax[rows.use,"調定済額標準税率超過調定額."] / actual.tax[rows.use,"調定済額現年課税分.A."]
  if(!is.na(boot)) actual.chou <- boot.chou[rows.use]
  if(!TAX.FLOOR) actual.chou <- actual.chou + b["tau.bar"]

  optimal.rate <- v.m.new(actual.tax[rows.use,colnames(m.dat)], b, time=1, ret="tax")
  optimal.recentered <- optimal.rate - b["tau.bar"]
  if(!TAX.FLOOR) optimal.recentered <- optimal.rate

  log.s <- log(actual.tax[rows.use,"sfn.struct.one"])
  creg.null <- censReg(actual.chou ~ optimal.recentered, start=c(0, 1, 0), fixed=c(TRUE, TRUE, FALSE))

  f.grad <- function(x) {
    lik.r <- censReg(actual.chou ~ optimal.recentered + log.s, start=c(x[1], creg.null$estimate[2], x[2], creg.null$estimate[3]), logLikOnly=TRUE)
    return(attr(lik.r, "gradient")[,c(1,3)])
  }
  U <- f.grad(c(0,0))
  neg.I <- -jacobian(function(x) colSums(f.grad(x)), c(0,0))

  return(list(U=U, neg.I=neg.I))
}

if(is.na(boot)) t1.rate <- c("14"=5, "13"=2, "27"=2)

if(!is.na(boot)) t1.rate <- c("14"=late[14], "13"=late[13], "27"=late[27])

edge.var <- vector("numeric",47)
for(p in prefs) {
  e.mat <- sapply(E(p.graph.l[[p]]), function(x) V(p.graph.l[[p]])[inc(x)]$name)
  pop.mat <- apply(e.mat, 1:2, function(x) d.h13[x, "pop"])
  edge.var[p] <- sum(1/apply(sqrt(pop.mat), 2, prod))
}



pen <- function(par) {
  print(par)
  b <- par.fixed
  b[names(par)] <- par

  if(b["B1"] < 0) b["B1"] <- - b["B1"]
  if(b["B0"] < 0) b["B0"] <- - b["B0"]
  
  tmp.r <- mclapply(prefs, function(i) f.r.l[[as.character(i)]](b))

  m1.v.i.r <- unlist(tmp.r)
  if(any(m1.v.i.r == Inf)) return(Inf)

  tmp.c <- mclapply(prefs, function(i) f.c.l[[as.character(i)]](b))
  m1.v.i.c <- sapply(tmp.c, function(x) sum(sapply(x, function(y) y$obj)))
  if(any(m1.v.i.c == Inf)) return(Inf)

  m1.v.i <- nmunis[prefs] + edge.var[prefs] - (m1.v.i.r + m1.v.i.c)

  m1.v <- sapply(pref.groups, function(pg) sum(m1.v.i[pg]))

  base.rate <- lapply(pref.groups, function(p) colSums(early[,p]) / nrow(early) * (2010 - 1999))
  m4.v <- rep(NA, 3)
  tmp.m <- mclapply(1:length(prefs), function(i) f.m.l[[i]](b, time=0))
  names(tmp.m) <- as.character(prefs)
  pg.munis <- sapply(pref.groups, function(x) sum(nmunis[x]))
  for(n in 1:length(pref.groups)) {
    p.group <- pref.groups[[n]]
    tmp.2 <- sapply(as.list(p.group), function(x) tmp.m[[as.character(x)]])
    ecdf.data <- sort(rowSums(tmp.2))
    obs <- sum(base.rate[[n]]) + 0.5  # continuity correction
    if(obs > max(ecdf.data)) {
      m4.v[n] <- Inf
    } else {
      ecdf.min <- sum(apply(tmp.2, 2, sort)[1,])
      approx.x <- c(0, ecdf.min, ecdf.data)
      approx.y <- c(0, nrow(tmp.2)^-ncol(tmp.2), (1:nrow(tmp.2))/nrow(tmp.2))
      tmp.qtile <- approx(x=approx.x, y=approx.y, xout=obs)$y
      m4.v[n] <- qnorm(tmp.qtile)
    }
  }

  p.group <- pref.groups[[1]]
  tmp.1 <- mclapply(pref.groups[[1]], function(i) f.m.l[[i]](b, time=1))
  names(tmp.1) <- as.character(pref.groups[[1]])
  tmp.2 <- sapply(as.list(p.group), function(x) tmp.1[[as.character(x)]])
  do.m5 <- function(p.idx) {
    ecdf.data <- sort(rowSums(tmp.2[,p.idx]))
    obs <- sum(t1.rate[p.idx]) + 0.5  # continuity correction
    if(obs > max(ecdf.data)) {
      m5 <- Inf
    } else {
      ecdf.min <- sum(apply(tmp.2[,p.idx], 2, sort)[1,])
      approx.x <- c(0, ecdf.min, ecdf.data)
      approx.y <- c(0, nrow(tmp.2)^-length(p.idx), (1:nrow(tmp.2))/nrow(tmp.2))
      tmp.qtile <- approx(x=approx.x, y=approx.y, xout=obs)$y
      m5 <- qnorm(tmp.qtile)
    }
    return(m5)
  }
  m5 <- do.m5(1:3)
  m5.subsample <- sapply(list(c(1,2), c(1,3), c(2,3)), do.m5)
  pg <- pref.groups[[1]]

  f <- function(p.nums) sum(m1.v.i.r[p.nums])/sqrt(2*sum(nmunis[p.nums] + edge.var[p.nums]))

  m1.1.subsample <- sapply(combn(pref.groups[[1]], 2, simplify=FALSE), f)
  m1.2.subsample <- sapply(combn(pref.groups[[2]], 2, simplify=FALSE), f)
  m1.3.subsample <- sapply(combn(pref.groups[[3]], 2, simplify=FALSE), f)
  
  m1.m5.cor <- cor(m1.1.subsample, m5.subsample, method="spearman")
  
  tmp.l <- pen.tax(prefs, b=b)
  D <- chol(solve(tmp.l$neg.I))
  m6.v.i <- tmp.l$U %*% t(D)
  m6.v <- colSums(m6.v.i)
  
  pg <- pref.groups[[1]]
  f <- function(p.nums) {
    rows.use <- actual.tax[,"都道府県コード"] %in% p.nums
    ret <- colSums(m6.v.i[rows.use,])/sqrt(sum(rows.use))
    return(ret)
  }
  m6.1.subsample <- sapply(combn(pref.groups[[1]], 2, simplify=FALSE), f)
  m6.2.subsample <- sapply(combn(pref.groups[[2]], 2, simplify=FALSE), f)
  m6.3.subsample <- sapply(combn(pref.groups[[3]], 2, simplify=FALSE), f)
  m1.m6.1.cor <- cor(m1.v.i, sapply(1:47, f)[1,])
  m1.m6.2.cor <- cor(m1.v.i, sapply(1:47, f)[2,])

  m5.m6.1.cor <- cor(m5.subsample, m6.1.subsample[1,])
  m5.m6.2.cor <- cor(m5.subsample, m6.1.subsample[2,])
  
  s.hat <- matrix(0, nrow=9, ncol=9)
  diag(s.hat)[1:3] <- sapply(pref.groups, function(pg) 2*sum(nmunis[pg] + edge.var[pg]))
  diag(s.hat)[4:7] <- 1

  diag(s.hat)[8:9] <- 1
  s.hat[1,7] <- s.hat[7,1] <- m1.m5.cor * sqrt(s.hat[1,1])

  s.hat[1,8] <- s.hat[8,1] <- m1.m6.1.cor * sqrt(s.hat[1,1]) * sqrt(pg.munis[1]/sum(pg.munis))
  s.hat[1,9] <- s.hat[9,1] <- m1.m6.2.cor * sqrt(s.hat[1,1]) * sqrt(pg.munis[1]/sum(pg.munis))

  s.hat[2,8] <- s.hat[8,2] <- m1.m6.1.cor * sqrt(s.hat[2,2]) * sqrt(pg.munis[2]/sum(pg.munis))
  s.hat[2,9] <- s.hat[9,2] <- m1.m6.2.cor * sqrt(s.hat[2,2]) * sqrt(pg.munis[2]/sum(pg.munis))

  s.hat[3,8] <- s.hat[8,3] <- m1.m6.1.cor * sqrt(s.hat[3,3]) * sqrt(pg.munis[3]/sum(pg.munis))
  s.hat[3,9] <- s.hat[9,3] <- m1.m6.2.cor * sqrt(s.hat[3,3]) * sqrt(pg.munis[3]/sum(pg.munis))
  s.hat[7,8] <- s.hat[8,7] <- m5.m6.1.cor * sqrt(pg.munis[1]/sum(pg.munis))
  s.hat[7,9] <- s.hat[9,7] <- m5.m6.2.cor * sqrt(pg.munis[1]/sum(pg.munis))
  e <- eigen(s.hat)  # estimating correlations may produce eigenvalues that are slightly negative
  s.hat.pd <- e$vectors %*% diag(pmax(e$values, 0.05)) %*% solve(e$vectors)  # ... so force the covariance matrix to be positive definite
  omega.hat <- diag(diag(s.hat.pd)^(-1/2)) %*% s.hat.pd %*% diag(diag(s.hat.pd)^(-1/2))

  old.seed <- .Random.seed
  set.seed(1)
  rnd.m <- rmvnorm(TEST.DRAWS, sigma=omega.hat)
  .Random.seed <<- old.seed
  rnd.S <- apply(rnd.m, 1:2, function(x) min(x, 0))^2
  qtiles <- c(.5, .8, .9, .95, .99, .999, .9999)
  if(USE.TAX) q.S <- quantile(rowSums(rnd.S), probs=qtiles)
  if(!USE.TAX) q.S <- quantile(rowSums(rnd.S[,-c(8,9)]), probs=qtiles)
  

  m <- c(pmin(0, c(m1.v, m4.v, m5)), m6.v)

  if(USE.TAX) m.pen <- m^2/diag(s.hat.pd)

  T <- sum(m.pen)
  return(T/q.S["95%"])
}

print(par.fixed)

## pen(par)
parscale <- c(B2=.1, B0=100, B1=10, B4=.1, B3=.2)
stopifnot(names(parscale) == names(par))
## names(parscale) <- NULL  # breaks optim for some reason

if(is.na(fixed.var)) {
  par.use <- par.best
  parscale.use <- parscale
} else {  # take out the variable that's fixed
  drop.idx <- which(fixed.var == names(par.best))
  par.use <- par.best[-drop.idx]
  parscale.use <- parscale[-drop.idx]
}


if(is.na(boot)) {
  print(pen(par.best)) # verify that calculations up to this point are deterministic (CPLEX introduces a very small amount of noise)
  set.seed(as.integer(proc.time()["elapsed"] * 1000))  # things become non-deterministic here
  ## par.rnd <- par.use
  par.rnd <- par.use*exp(rnorm(length(par.use), sd=1/4))  # random starting point (.../2 is too big)
  ## par.use <- par.best
  print(par.rnd)
  pen(par.use)
  nl.out <- optim(par.rnd, fn=pen, control=list(trace=6, parscale=parscale.use, maxit=10000))
  ## nl.out <- optim(par.use, fn=pen, control=list(trace=6, parscale=parscale.use, maxit=10000))
  nl.out
} else {
  par.boot <- par.best
  b <- par.fixed
  b[names(par.boot)] <- par.boot
  boot.rnd <- rnorm(nrow(actual.tax), sd=exp(-4.58694))
  boot.tau.opt <- v.m.new(actual.tax[,colnames(m.dat)], b, time=1, ret="tax")
  boot.chou <- pmax(0, boot.tau.opt + boot.rnd - b["tau.bar"])
  pen(par.boot)
}

